@@ -21,12 +21,15 @@
  * OF THIS SOFTWARE.
  */
 
+#include "config.h"
+
 #include <stdint.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
 #include <fcntl.h>
 #include <unistd.h>
+#include <errno.h>
 #include <math.h>
 #include <cairo.h>
 #include <sys/wait.h>
@@ -49,6 +52,7 @@ extern char **environ; /* defined by libc */
 struct desktop {
 	struct display *display;
 	struct desktop_shell *shell;
+	uint32_t interface_version;
 	struct unlock_dialog *unlock_dialog;
 	struct task unlock_task;
 	struct wl_list outputs;
@@ -56,7 +60,12 @@ struct desktop {
 	struct window *grab_window;
 	struct widget *grab_widget;
 
+	struct weston_config *config;
+	int locking;
+
 	enum cursor_type grab_cursor;
+
+	int painted;
 };
 
 struct surface {
@@ -72,16 +81,24 @@ struct panel {
 	struct widget *widget;
 	struct wl_list launcher_list;
 	struct panel_clock *clock;
+	int painted;
+	uint32_t color;
 };
 
 struct background {
 	struct surface base;
 	struct window *window;
 	struct widget *widget;
+	int painted;
+
+	char *image;
+	int type;
+	uint32_t color;
 };
 
 struct output {
 	struct wl_output *output;
+	uint32_t server_output_id;
 	struct wl_list link;
 
 	struct panel *panel;
@@ -115,35 +132,8 @@ struct unlock_dialog {
 	struct desktop *desktop;
 };
 
-static char *key_background_image = DATADIR "/weston/pattern.png";
-static char *key_background_type = "tile";
-static uint32_t key_panel_color = 0xaa000000;
-static uint32_t key_background_color = 0xff002244;
-static char *key_launcher_icon;
-static char *key_launcher_path;
-static void launcher_section_done(void *data);
-static int key_locking = 1;
-
-static const struct config_key shell_config_keys[] = {
-	{ "background-image", CONFIG_KEY_STRING, &key_background_image },
-	{ "background-type", CONFIG_KEY_STRING, &key_background_type },
-	{ "panel-color", CONFIG_KEY_UNSIGNED_INTEGER, &key_panel_color },
-	{ "background-color", CONFIG_KEY_UNSIGNED_INTEGER, &key_background_color },
-	{ "locking", CONFIG_KEY_BOOLEAN, &key_locking },
-};
-
-static const struct config_key launcher_config_keys[] = {
-	{ "icon", CONFIG_KEY_STRING, &key_launcher_icon },
-	{ "path", CONFIG_KEY_STRING, &key_launcher_path },
-};
-
-static const struct config_section config_sections[] = {
-	{ "shell",
-	  shell_config_keys, ARRAY_LENGTH(shell_config_keys) },
-	{ "launcher",
-	  launcher_config_keys, ARRAY_LENGTH(launcher_config_keys),
-	  launcher_section_done }
-};
+static void
+panel_add_launchers(struct panel *panel, struct desktop *desktop);
 
 static void
 sigchild_handler(int s)
@@ -156,7 +146,7 @@ sigchild_handler(int s)
 }
 
 static void
-menu_func(struct window *window, int index, void *data)
+menu_func(void *data, struct input *input, int index)
 {
 	printf("Selected index %d from a panel menu.\n", index);
 }
@@ -175,6 +165,38 @@ show_menu(struct panel *panel, struct input *input, uint32_t time)
 			 x - 10, y - 10, menu_func, entries, 4);
 }
 
+static int
+is_desktop_painted(struct desktop *desktop)
+{
+	struct output *output;
+
+	wl_list_for_each(output, &desktop->outputs, link) {
+		if (output->panel && !output->panel->painted)
+			return 0;
+		if (output->background && !output->background->painted)
+			return 0;
+	}
+
+	return 1;
+}
+
+static void
+check_desktop_ready(struct window *window)
+{
+	struct display *display;
+	struct desktop *desktop;
+
+	display = window_get_display(window);
+	desktop = display_get_user_data(display);
+
+	if (!desktop->painted && is_desktop_painted(desktop)) {
+		desktop->painted = 1;
+
+		if (desktop->interface_version >= 2)
+			desktop_shell_desktop_ready(desktop->shell);
+	}
+}
+
 static void
 panel_launcher_activate(struct panel_launcher *widget)
 {
@@ -201,12 +223,10 @@ static void
 panel_launcher_redraw_handler(struct widget *widget, void *data)
 {
 	struct panel_launcher *launcher = data;
-	cairo_surface_t *surface;
 	struct rectangle allocation;
 	cairo_t *cr;
 
-	surface = window_get_surface(launcher->panel->window);
-	cr = cairo_create(surface);
+	cr = widget_cairo_create(launcher->panel->widget);
 
 	widget_get_allocation(widget, &allocation);
 	if (launcher->pressed) {
@@ -255,14 +275,16 @@ panel_redraw_handler(struct widget *widget, void *data)
 	cairo_t *cr;
 	struct panel *panel = data;
 
-	surface = window_get_surface(panel->window);
-	cr = cairo_create(surface);
+	cr = widget_cairo_create(panel->widget);
 	cairo_set_operator(cr, CAIRO_OPERATOR_SOURCE);
-	set_hex_color(cr, key_panel_color);
+	set_hex_color(cr, panel->color);
 	cairo_paint(cr);
 
 	cairo_destroy(cr);
+	surface = window_get_surface(panel->window);
 	cairo_surface_destroy(surface);
+	panel->painted = 1;
+	check_desktop_ready(panel->window);
 }
 
 static int
@@ -300,6 +322,32 @@ panel_launcher_button_handler(struct widget *widget,
 	widget_schedule_redraw(widget);
 	if (state == WL_POINTER_BUTTON_STATE_RELEASED)
 		panel_launcher_activate(launcher);
+
+}
+
+static void
+panel_launcher_touch_down_handler(struct widget *widget, struct input *input,
+				  uint32_t serial, uint32_t time, int32_t id,
+				  float x, float y, void *data)
+{
+	struct panel_launcher *launcher;
+
+	launcher = widget_get_user_data(widget);
+	launcher->focused = 1;
+	widget_schedule_redraw(widget);
+}
+
+static void
+panel_launcher_touch_up_handler(struct widget *widget, struct input *input,
+				uint32_t serial, uint32_t time, int32_t id, 
+				void *data)
+{
+	struct panel_launcher *launcher;
+
+	launcher = widget_get_user_data(widget);
+	launcher->focused = 0;
+	widget_schedule_redraw(widget);
+	panel_launcher_activate(launcher);
 }
 
 static void
@@ -317,7 +365,6 @@ clock_func(struct task *task, uint32_t events)
 static void
 panel_clock_redraw_handler(struct widget *widget, void *data)
 {
-	cairo_surface_t *surface;
 	struct panel_clock *clock = data;
 	cairo_t *cr;
 	struct rectangle allocation;
@@ -335,8 +382,7 @@ panel_clock_redraw_handler(struct widget *widget, void *data)
 	if (allocation.width == 0)
 		return;
 
-	surface = window_get_surface(clock->panel->window);
-	cr = cairo_create(surface);
+	cr = widget_cairo_create(clock->panel->widget);
 	cairo_select_font_face(cr, "sans",
 			       CAIRO_FONT_SLANT_NORMAL,
 			       CAIRO_FONT_WEIGHT_NORMAL);
@@ -393,8 +439,7 @@ panel_add_clock(struct panel *panel)
 		return;
 	}
 
-	clock = malloc(sizeof *clock);
-	memset(clock, 0, sizeof *clock);
+	clock = xzalloc(sizeof *clock);
 	clock->panel = panel;
 	panel->clock = clock;
 	clock->clock_fd = timerfd;
@@ -491,15 +536,15 @@ panel_destroy(struct panel *panel)
 }
 
 static struct panel *
-panel_create(struct display *display)
+panel_create(struct desktop *desktop)
 {
 	struct panel *panel;
+	struct weston_config_section *s;
 
-	panel = malloc(sizeof *panel);
-	memset(panel, 0, sizeof *panel);
+	panel = xzalloc(sizeof *panel);
 
 	panel->base.configure = panel_configure;
-	panel->window = window_create_custom(display);
+	panel->window = window_create_custom(desktop->display);
 	panel->widget = window_add_widget(panel->window, panel);
 	wl_list_init(&panel->launcher_list);
 
@@ -512,6 +557,12 @@ panel_create(struct display *display)
 	
 	panel_add_clock(panel);
 
+	s = weston_config_get_section(desktop->config, "shell", NULL, NULL);
+	weston_config_section_get_uint(s, "panel-color",
+				       &panel->color, 0xaa000000);
+
+	panel_add_launchers(panel, desktop);
+
 	return panel;
 }
 
@@ -519,13 +570,16 @@ static cairo_surface_t *
 load_icon_or_fallback(const char *icon)
 {
 	cairo_surface_t *surface = cairo_image_surface_create_from_png(icon);
+	cairo_status_t status;
 	cairo_t *cr;
 
-	if (cairo_surface_status(surface) == CAIRO_STATUS_SUCCESS)
+	status = cairo_surface_status(surface);
+	if (status == CAIRO_STATUS_SUCCESS)
 		return surface;
 
 	cairo_surface_destroy(surface);
-	fprintf(stderr, "ERROR loading icon from file '%s'\n", icon);
+	fprintf(stderr, "ERROR loading icon from file '%s', error: '%s'\n",
+		icon, cairo_status_to_string(status));
 
 	/* draw fallback icon */
 	surface = cairo_image_surface_create(CAIRO_FORMAT_ARGB32,
@@ -556,10 +610,9 @@ panel_add_launcher(struct panel *panel, const char *icon, const char *path)
 	char *start, *p, *eq, **ps;
 	int i, j, k;
 
-	launcher = malloc(sizeof *launcher);
-	memset(launcher, 0, sizeof *launcher);
+	launcher = xzalloc(sizeof *launcher);
 	launcher->icon = load_icon_or_fallback(icon);
-	launcher->path = strdup(path);
+	launcher->path = xstrdup(path);
 
 	wl_array_init(&launcher->envp);
 	wl_array_init(&launcher->argv);
@@ -614,6 +667,10 @@ panel_add_launcher(struct panel *panel, const char *icon, const char *path)
 				   panel_launcher_leave_handler);
 	widget_set_button_handler(launcher->widget,
 				    panel_launcher_button_handler);
+	widget_set_touch_down_handler(launcher->widget,
+				      panel_launcher_touch_down_handler);
+	widget_set_touch_up_handler(launcher->widget,
+				    panel_launcher_touch_up_handler);
 	widget_set_redraw_handler(launcher->widget,
 				  panel_launcher_redraw_handler);
 	widget_set_motion_handler(launcher->widget,
@@ -622,6 +679,7 @@ panel_add_launcher(struct panel *panel, const char *icon, const char *path)
 
 enum {
 	BACKGROUND_SCALE,
+	BACKGROUND_SCALE_CROP,
 	BACKGROUND_TILE
 };
 
@@ -633,64 +691,67 @@ background_draw(struct widget *widget, void *data)
 	cairo_pattern_t *pattern;
 	cairo_matrix_t matrix;
 	cairo_t *cr;
-	double sx, sy;
+	double im_w, im_h;
+	double sx, sy, s;
+	double tx, ty;
 	struct rectangle allocation;
-	int type = -1;
-	struct display *display;
-	struct wl_region *opaque;
 
 	surface = window_get_surface(background->window);
 
-	cr = cairo_create(surface);
+	cr = widget_cairo_create(background->widget);
 	cairo_set_operator(cr, CAIRO_OPERATOR_SOURCE);
 	cairo_set_source_rgba(cr, 0.0, 0.0, 0.2, 1.0);
 	cairo_paint(cr);
 
 	widget_get_allocation(widget, &allocation);
 	image = NULL;
-	if (key_background_image)
-		image = load_cairo_surface(key_background_image);
+	if (background->image)
+		image = load_cairo_surface(background->image);
+	else if (background->color == 0)
+		image = load_cairo_surface(DATADIR "/weston/pattern.png");
 
-	if (strcmp(key_background_type, "scale") == 0)
-		type = BACKGROUND_SCALE;
-	else if (strcmp(key_background_type, "tile") == 0)
-		type = BACKGROUND_TILE;
-	else
-		fprintf(stderr, "invalid background-type: %s\n",
-			key_background_type);
+	if (image && background->type != -1) {
+		im_w = cairo_image_surface_get_width(image);
+		im_h = cairo_image_surface_get_height(image);
+		sx = im_w / allocation.width;
+		sy = im_h / allocation.height;
 
-	if (image && type != -1) {
 		pattern = cairo_pattern_create_for_surface(image);
-		switch (type) {
+
+		switch (background->type) {
 		case BACKGROUND_SCALE:
-			sx = (double) cairo_image_surface_get_width(image) /
-				allocation.width;
-			sy = (double) cairo_image_surface_get_height(image) /
-				allocation.height;
 			cairo_matrix_init_scale(&matrix, sx, sy);
 			cairo_pattern_set_matrix(pattern, &matrix);
+			cairo_pattern_set_extend(pattern, CAIRO_EXTEND_PAD);
+			break;
+		case BACKGROUND_SCALE_CROP:
+			s = (sx < sy) ? sx : sy;
+			/* align center */
+			tx = (im_w - s * allocation.width) * 0.5;
+			ty = (im_h - s * allocation.height) * 0.5;
+			cairo_matrix_init_translate(&matrix, tx, ty);
+			cairo_matrix_scale(&matrix, s, s);
+			cairo_pattern_set_matrix(pattern, &matrix);
+			cairo_pattern_set_extend(pattern, CAIRO_EXTEND_PAD);
 			break;
 		case BACKGROUND_TILE:
 			cairo_pattern_set_extend(pattern, CAIRO_EXTEND_REPEAT);
 			break;
 		}
+
 		cairo_set_source(cr, pattern);
 		cairo_pattern_destroy (pattern);
 		cairo_surface_destroy(image);
 	} else {
-		set_hex_color(cr, key_background_color);
+		set_hex_color(cr, background->color);
 	}
 
 	cairo_paint(cr);
 	cairo_destroy(cr);
 	cairo_surface_destroy(surface);
 
-	display = window_get_display(background->window);
-	opaque = wl_compositor_create_region(display_get_compositor(display));
-	wl_region_add(opaque, allocation.x, allocation.y,
-		      allocation.width, allocation.height);
-	wl_surface_set_opaque_region(window_get_wl_surface(background->window), opaque);
-	wl_region_destroy(opaque);
+	background->painted = 1;
+	check_desktop_ready(background->window);
 }
 
 static void
@@ -710,13 +771,12 @@ unlock_dialog_redraw_handler(struct widget *widget, void *data)
 {
 	struct unlock_dialog *dialog = data;
 	struct rectangle allocation;
-	cairo_t *cr;
 	cairo_surface_t *surface;
+	cairo_t *cr;
 	cairo_pattern_t *pat;
 	double cx, cy, r, f;
 
-	surface = window_get_surface(dialog->window);
-	cr = cairo_create(surface);
+	cr = widget_cairo_create(widget);
 
 	widget_get_allocation(dialog->widget, &allocation);
 	cairo_rectangle(cr, allocation.x, allocation.y,
@@ -749,6 +809,7 @@ unlock_dialog_redraw_handler(struct widget *widget, void *data)
 
 	cairo_destroy(cr);
 
+	surface = window_get_surface(dialog->window);
 	cairo_surface_destroy(surface);
 }
 
@@ -770,6 +831,31 @@ unlock_dialog_button_handler(struct widget *widget,
 	}
 }
 
+static void
+unlock_dialog_touch_down_handler(struct widget *widget, struct input *input,
+		   uint32_t serial, uint32_t time, int32_t id,
+		   float x, float y, void *data)
+{
+	struct unlock_dialog *dialog = data;
+
+	dialog->button_focused = 1;
+	widget_schedule_redraw(widget);
+}
+
+static void
+unlock_dialog_touch_up_handler(struct widget *widget, struct input *input,
+				uint32_t serial, uint32_t time, int32_t id,
+				void *data)
+{
+	struct unlock_dialog *dialog = data;
+	struct desktop *desktop = dialog->desktop;
+
+	dialog->button_focused = 0;
+	widget_schedule_redraw(widget);
+	display_defer(desktop->display, &desktop->unlock_task);
+	dialog->closing = 1;
+}
+
 static void
 unlock_dialog_keyboard_focus_handler(struct window *window,
 				     struct input *device, void *data)
@@ -806,13 +892,10 @@ unlock_dialog_create(struct desktop *desktop)
 	struct display *display = desktop->display;
 	struct unlock_dialog *dialog;
 
-	dialog = malloc(sizeof *dialog);
-	if (!dialog)
-		return NULL;
-	memset(dialog, 0, sizeof *dialog);
+	dialog = xzalloc(sizeof *dialog);
 
 	dialog->window = window_create_custom(display);
-	dialog->widget = frame_create(dialog->window, dialog);
+	dialog->widget = window_frame_create(dialog->window, dialog);
 	window_set_title(dialog->window, "Unlock your desktop");
 
 	window_set_user_data(dialog->window, dialog);
@@ -827,6 +910,10 @@ unlock_dialog_create(struct desktop *desktop)
 				 unlock_dialog_widget_leave_handler);
 	widget_set_button_handler(dialog->button,
 				  unlock_dialog_button_handler);
+	widget_set_touch_down_handler(dialog->button,
+				      unlock_dialog_touch_down_handler);
+	widget_set_touch_up_handler(dialog->button,
+				      unlock_dialog_touch_up_handler);
 
 	desktop_shell_set_lock_surface(desktop->shell,
 				       window_get_wl_surface(dialog->window));
@@ -873,7 +960,7 @@ desktop_shell_prepare_lock_surface(void *data,
 {
 	struct desktop *desktop = data;
 
-	if (!key_locking) {
+	if (!desktop->locking) {
 		desktop_shell_unlock(desktop->shell);
 		return;
 	}
@@ -943,6 +1030,7 @@ background_destroy(struct background *background)
 	widget_destroy(background->widget);
 	window_destroy(background->window);
 
+	free(background->image);
 	free(background);
 }
 
@@ -950,15 +1038,45 @@ static struct background *
 background_create(struct desktop *desktop)
 {
 	struct background *background;
+	struct weston_config_section *s;
+	char *type;
 
-	background = malloc(sizeof *background);
-	memset(background, 0, sizeof *background);
-
+	background = xzalloc(sizeof *background);
 	background->base.configure = background_configure;
 	background->window = window_create_custom(desktop->display);
 	background->widget = window_add_widget(background->window, background);
 	window_set_user_data(background->window, background);
 	widget_set_redraw_handler(background->widget, background_draw);
+	widget_set_transparent(background->widget, 0);
+	window_set_preferred_format(background->window,
+				    WINDOW_PREFERRED_FORMAT_RGB565);
+
+	s = weston_config_get_section(desktop->config, "shell", NULL, NULL);
+	weston_config_section_get_string(s, "background-image",
+					 &background->image, NULL);
+	weston_config_section_get_uint(s, "background-color",
+				       &background->color, 0);
+
+	weston_config_section_get_string(s, "background-type",
+					 &type, "tile");
+	if (type == NULL) {
+		fprintf(stderr, "%s: out of memory\n", program_invocation_short_name);
+		exit(EXIT_FAILURE);
+	}
+
+	if (strcmp(type, "scale") == 0) {
+		background->type = BACKGROUND_SCALE;
+	} else if (strcmp(type, "scale-crop") == 0) {
+		background->type = BACKGROUND_SCALE_CROP;
+	} else if (strcmp(type, "tile") == 0) {
+		background->type = BACKGROUND_TILE;
+	} else {
+		background->type = -1;
+		fprintf(stderr, "invalid background-type: %s\n",
+			type);
+	}
+
+	free(type);
 
 	return background;
 }
@@ -1004,7 +1122,8 @@ static void
 output_destroy(struct output *output)
 {
 	background_destroy(output->background);
-	panel_destroy(output->panel);
+	if (output->panel)
+		panel_destroy(output->panel);
 	wl_output_destroy(output->output);
 	wl_list_remove(&output->link);
 
@@ -1021,6 +1140,96 @@ desktop_destroy_outputs(struct desktop *desktop)
 		output_destroy(output);
 }
 
+static void
+output_handle_geometry(void *data,
+                       struct wl_output *wl_output,
+                       int x, int y,
+                       int physical_width,
+                       int physical_height,
+                       int subpixel,
+                       const char *make,
+                       const char *model,
+                       int transform)
+{
+	struct output *output = data;
+
+	if (output->panel)
+		window_set_buffer_transform(output->panel->window, transform);
+	window_set_buffer_transform(output->background->window, transform);
+}
+
+static void
+output_handle_mode(void *data,
+		   struct wl_output *wl_output,
+		   uint32_t flags,
+		   int width,
+		   int height,
+		   int refresh)
+{
+}
+
+static void
+output_handle_done(void *data,
+                   struct wl_output *wl_output)
+{
+}
+
+static void
+output_handle_scale(void *data,
+                    struct wl_output *wl_output,
+                    int32_t scale)
+{
+	struct output *output = data;
+
+	if (output->panel)
+		window_set_buffer_scale(output->panel->window, scale);
+	window_set_buffer_scale(output->background->window, scale);
+}
+
+static const struct wl_output_listener output_listener = {
+	output_handle_geometry,
+	output_handle_mode,
+	output_handle_done,
+	output_handle_scale
+};
+
+static int
+want_panel(struct desktop *desktop)
+{
+	struct weston_config_section *s;
+	char *location = NULL;
+	int ret = 1;
+
+	s = weston_config_get_section(desktop->config, "shell", NULL, NULL);
+	weston_config_section_get_string(s, "panel-location",
+					 &location, "top");
+
+	if (strcmp(location, "top") != 0)
+		ret = 0;
+
+	free(location);
+
+	return ret;
+}
+
+static void
+output_init(struct output *output, struct desktop *desktop)
+{
+	struct wl_surface *surface;
+
+	if (want_panel(desktop)) {
+		output->panel = panel_create(desktop);
+		surface = window_get_wl_surface(output->panel->window);
+		desktop_shell_set_panel(desktop->shell,
+					output->output, surface);
+	}
+
+	output->background = background_create(desktop);
+	surface = window_get_wl_surface(output->background->window);
+	desktop_shell_set_background(desktop->shell,
+				     output->output, surface);
+}
+
 static void
 create_output(struct desktop *desktop, uint32_t id)
 {
@@ -1031,9 +1240,17 @@ create_output(struct desktop *desktop, uint32_t id)
 		return;
 
 	output->output =
-		display_bind(desktop->display, id, &wl_output_interface, 1);
+		display_bind(desktop->display, id, &wl_output_interface, 2);
+	output->server_output_id = id;
+
+	wl_output_add_listener(output->output, &output_listener, output);
 
 	wl_list_insert(&desktop->outputs, &output->link);
+
+	/* On start up we may process an output global before the shell global
+	 * in which case we can't create the panel and background just yet */
+	if (desktop->shell)
+		output_init(output, desktop);
 }
 
 static void
@@ -1043,8 +1260,10 @@ global_handler(struct display *display, uint32_t id,
 	struct desktop *desktop = data;
 
 	if (!strcmp(interface, "desktop_shell")) {
+		desktop->interface_version = (version < 2) ? version : 2;
 		desktop->shell = display_bind(desktop->display,
-					      id, &desktop_shell_interface, 1);
+					      id, &desktop_shell_interface,
+					      desktop->interface_version);
 		desktop_shell_add_listener(desktop->shell, &listener, desktop);
 	} else if (!strcmp(interface, "wl_output")) {
 		create_output(desktop, id);
@@ -1052,49 +1271,72 @@ global_handler(struct display *display, uint32_t id,
 }
 
 static void
-launcher_section_done(void *data)
+global_handler_remove(struct display *display, uint32_t id,
+	       const char *interface, uint32_t version, void *data)
 {
 	struct desktop *desktop = data;
 	struct output *output;
 
-	if (key_launcher_icon == NULL || key_launcher_path == NULL) {
-		fprintf(stderr, "invalid launcher section\n");
-		return;
+	if (!strcmp(interface, "wl_output")) {
+		wl_list_for_each(output, &desktop->outputs, link) {
+			if (output->server_output_id == id) {
+				output_destroy(output);
+				break;
+			}
+		}
 	}
-
-	wl_list_for_each(output, &desktop->outputs, link) {
-		panel_add_launcher(output->panel,
-				   key_launcher_icon, key_launcher_path);
-	}
-
-	free(key_launcher_icon);
-	key_launcher_icon = NULL;
-	free(key_launcher_path);
-	key_launcher_path = NULL;
 }
 
 static void
-add_default_launcher(struct desktop *desktop)
+panel_add_launchers(struct panel *panel, struct desktop *desktop)
 {
-	struct output *output;
+	struct weston_config_section *s;
+	char *icon, *path;
+	const char *name;
+	int count;
 
-	wl_list_for_each(output, &desktop->outputs, link)
-		panel_add_launcher(output->panel,
+	count = 0;
+	s = NULL;
+	while (weston_config_next_section(desktop->config, &s, &name)) {
+		if (strcmp(name, "launcher") != 0)
+			continue;
+
+		weston_config_section_get_string(s, "icon", &icon, NULL);
+		weston_config_section_get_string(s, "path", &path, NULL);
+
+		if (icon != NULL && path != NULL) {
+			panel_add_launcher(panel, icon, path);
+			count++;
+		} else {
+			fprintf(stderr, "invalid launcher section\n");
+		}
+
+		free(icon);
+		free(path);
+	}
+
+	if (count == 0) {
+		/* add default launcher */
+		panel_add_launcher(panel,
 				   DATADIR "/weston/terminal.png",
 				   BINDIR "/weston-terminal");
+	}
 }
 
 int main(int argc, char *argv[])
 {
 	struct desktop desktop = { 0 };
-	char *config_file;
 	struct output *output;
-	int ret;
+	struct weston_config_section *s;
 
 	desktop.unlock_task.run = unlock_dialog_finish;
 	wl_list_init(&desktop.outputs);
 
-	desktop.display = display_create(argc, argv);
+	desktop.config = weston_config_parse("weston.ini");
+	s = weston_config_get_section(desktop.config, "shell", NULL, NULL);
+	weston_config_section_get_bool(s, "locking", &desktop.locking, 1);
+
+	desktop.display = display_create(&argc, argv);
 	if (desktop.display == NULL) {
 		fprintf(stderr, "failed to create display: %m\n");
 		return -1;
@@ -1102,31 +1344,16 @@ int main(int argc, char *argv[])
 
 	display_set_user_data(desktop.display, &desktop);
 	display_set_global_handler(desktop.display, global_handler);
+	display_set_global_handler_remove(desktop.display, global_handler_remove);
 
-	wl_list_for_each(output, &desktop.outputs, link) {
-		struct wl_surface *surface;
-
-		output->panel = panel_create(desktop.display);
-		surface = window_get_wl_surface(output->panel->window);
-		desktop_shell_set_panel(desktop.shell,
-					output->output, surface);
-
-		output->background = background_create(&desktop);
-		surface = window_get_wl_surface(output->background->window);
-		desktop_shell_set_background(desktop.shell,
-					     output->output, surface);
-	}
+	/* Create panel and background for outputs processed before the shell
+	 * global interface was processed */
+	wl_list_for_each(output, &desktop.outputs, link)
+		if (!output->panel)
+			output_init(output, &desktop);
 
 	grab_surface_create(&desktop);
 
-	config_file = config_file_path("weston.ini");
-	ret = parse_config_file(config_file,
-				config_sections, ARRAY_LENGTH(config_sections),
-				&desktop);
-	free(config_file);
-	if (ret < 0)
-		add_default_launcher(&desktop);
-
 	signal(SIGCHLD, sigchild_handler);
 
 	display_run(desktop.display);
