@@ -66,7 +66,7 @@ show_input_panel_surface(struct input_panel_surface *ipsurf)
 	float x, y;
 
 	wl_list_for_each(seat, &shell->compositor->seat_list, link) {
-		if (!seat->keyboard)
+		if (!seat->keyboard || !seat->keyboard->focus)
 			continue;
 		focus = weston_surface_get_main_surface(seat->keyboard->focus);
 		ipsurf->output = focus->output;
